// -*- C++ -*-


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/time.h>


static
double
now()
{
  timeval t;
  gettimeofday(&t, 0);
  return (t.tv_sec+1e-6*t.tv_usec);
}


struct context
{
  unsigned char b: 1;
  unsigned char em: 1;
  unsigned char i: 1;
  unsigned char s: 1;
  unsigned char tt: 1;
  unsigned char u: 2;
  unsigned char size: 4;
  unsigned char color: 4;
};


static
int
operator==(context const &decoration1, context const &decoration2)
{
  return (decoration1.b == decoration2.b &&
	  decoration1.em == decoration2.em &&
	  decoration1.i == decoration2.i &&
	  decoration1.s == decoration2.s &&
	  decoration1.tt == decoration2.tt &&
	  decoration1.u == decoration2.u &&
	  decoration1.size == decoration2.size &&
	  decoration1.color == decoration2.color);
}


static
int
operator!=(context const &decoration1, context const &decoration2)
{
  return (!(decoration1 == decoration2));
}


static
void
do_B(context &decoration)
{
  decoration.b=1;
}


static
void
do_EM(context &decoration)
{
  if (!decoration.s)
    decoration.em=!decoration.em;
}


static
void
do_I(context &decoration)
{
  decoration.i=1;
}


static
void
do_PL(context &decoration)
{
  decoration.s=0;
  decoration.em=0;
  decoration.i=0;
  decoration.b=0;
  decoration.tt=0;
  decoration.u=0;
}


static
void
do_S(context &decoration)
{
  decoration.s=1;
  decoration.em=0;
}


static
void
do_TT(context &decoration)
{
  decoration.tt=1;
}


static
void
do_U(context &decoration)
{
  if (decoration.u < 3)
    ++decoration.u;
}


static
void
do_0(context &decoration)
{
  decoration.size=0;
}


static
void
do_1(context &decoration)
{
  decoration.size=1;
}


static
void
do_2(context &decoration)
{
  decoration.size=2;
}


static
void
do_3(context &decoration)
{
  decoration.size=3;
}


static
void
do_4(context &decoration)
{
  decoration.size=4;
}


static
void
do_5(context &decoration)
{
  decoration.size=5;
}


static
void
do_6(context &decoration)
{
  decoration.size=6;
}


static
void
do_7(context &decoration)
{
  decoration.size=7;
}


static
void
do_8(context &decoration)
{
  decoration.size=8;
}


static
void
do_9(context &decoration)
{
  decoration.size=9;
}


static
void
do_r(context &decoration)
{
  decoration.color=0;
}


static
void
do_g(context &decoration)
{
  decoration.color=1;
}


static
void
do_b(context &decoration)
{
  decoration.color=2;
}


static
void
do_c(context &decoration)
{
  decoration.color=3;
}


static
void
do_m(context &decoration)
{
  decoration.color=4;
}


static
void
do_y(context &decoration)
{
  decoration.color=5;
}


static
void
do_k(context &decoration)
{
  decoration.color=6;
}


static
void
do_w(context &decoration)
{
  decoration.color=7;
}


static
context
space_context(context d)
{
  context decoration=d;
  decoration.s=0;
  decoration.em=0;
  decoration.i=0;
  decoration.b=0;
  if (decoration.u == 0)
    do_w(decoration);
  return (decoration);
}


static int const initial=15;


static
void
initialize(context &decoration)
{
  decoration.u=0;
  decoration.b=0;
  decoration.em=0;
  decoration.i=0;
  decoration.s=0;
  decoration.tt=0;
  decoration.size=initial;
  decoration.color=initial;
}


struct tag_info
{
  char const *tag;
  void (*do_it)(context &decoration);
};


static tag_info tB={ "B", do_B };
static tag_info tEM={ "EM", do_EM };
static tag_info tI={ "I", do_I };
static tag_info tPL={ "PL", do_PL };
static tag_info tS={ "S", do_S };
static tag_info tTT={ "TT", do_TT };
static tag_info tU={ "U", do_U };
static tag_info t0={ "0", do_0 };
static tag_info t1={ "1", do_1 };
static tag_info t2={ "2", do_2 };
static tag_info t3={ "3", do_3 };
static tag_info t4={ "4", do_4 };
static tag_info t5={ "5", do_5 };
static tag_info t6={ "6", do_6 };
static tag_info t7={ "7", do_7 };
static tag_info t8={ "8", do_8 };
static tag_info t9={ "9", do_9 };
static tag_info tr={ "r", do_r };
static tag_info tg={ "g", do_g };
static tag_info tb={ "b", do_b };
static tag_info tc={ "c", do_c };
static tag_info tm={ "m", do_m };
static tag_info ty={ "y", do_y };
static tag_info tk={ "k", do_k };
static tag_info tw={ "w", do_w };
static tag_info tinitialize={ 0, initialize };


static tag_info *tags[]=
{
  &tB,
  &tEM,
  &tI,
  &tPL,
  &tS,
  &tTT,
  &tU,
  &t0,
  &t1,
  &t2,
  &t3,
  &t4,
  &t5,
  &t6,
  &t7,
  &t8,
  &t9,
  &tr,
  &tg,
  &tb,
  &tc,
  &tm,
  &ty,
  &tk,
  &tw,
};


static tag_info *size_tags[]=
{
  &t0,
  &t1,
  &t2,
  &t3,
  &t4,
  &t5,
  &t6,
  &t7,
  &t8,
  &t9,
};


static tag_info *color_tags[]=
{
  &tr,
  &tg,
  &tb,
  &tc,
  &tm,
  &ty,
  &tk,
  &tw,
};



static
int
cost(tag_info *tag)
{
  return (2*strlen(tag->tag)+5);
}


static int num_s;


struct context_stack
{
  int refs;
  context decoration;
  tag_info *tag;
  context_stack *next;

#ifdef CHECK_ALLOC
  context_stack()
  {
    ++num_s;
  }

  ~context_stack()
  {
    --num_s;
    assert(num_s >= 0);
  }
#endif
};


static
context_stack *
copy(context_stack *s)
{
  if (s)
    ++s->refs;
  return (s);
}


static
void
free(context_stack *s)
{
  while (s && !--s->refs)
    {
      context_stack *next=s->next;
      delete s;
      s=next;
    }
}


static
context_stack *
push(tag_info *tag, context_stack *s)
{
  context_stack *result=new context_stack;
  result->refs=1;
  result->tag=tag;
  result->next=s;
  if (s)
    {
      result->decoration=s->decoration;
      assert(s->refs > 0);
    }
  return (result);
}


static
context_stack *
cow(context_stack *s)
{
  if (!s || s->refs == 1)
    return (s);
  --s->refs;
  return (push(s->tag, copy(s->next)));
}


static
context_stack *
pop(context_stack *s)
{
  assert(s);
  context_stack *result=copy(s->next);
  free(s);
  return (result);
}


static
context_stack *
do_it(context_stack *s, tag_info *tag)
{
  s=cow(push(tag, s));
  assert(s->refs == 1);
  (*tag->do_it)(s->decoration);
  return (s);
}


static char *input;


static
void
snarf()
{
  int max_len=1;
  input=(char *)malloc(max_len);
  int len=0;
  int c;
  while ((c=getchar()) != EOF)
    {
      if (len == max_len-1)
	{
	  max_len*=2;
	  input=(char *)realloc(input, max_len);
	}
      input[len++]=c;
    }
  input[len]='\0';
}


static
int
is_space(char c)
{
  return (c == ' ' || c == '\t' || c == '\n' || c == '\r');
}


struct decorated_char
{
  context decoration;
  char c;
};


static decorated_char *meaning;
static int meaning_len;


static
void
output(decorated_char *&m, char c, context const &decoration)
{
  m->decoration=decoration;
  m->c=c;
  ++m;
}


static
void
compute_meaning(char const *p)
{
  meaning=new decorated_char [strlen(p)+1];
  context_stack *s=do_it((context_stack *)0, &tinitialize);
  decorated_char *m=meaning;
  while (*p)
    if (*p == '<')
      {
	++p;
	int closing=*p == '/';
	if (closing)
	  ++p;
	char the_tag[100];
	char *t=the_tag;
	while (*p != '>')
	  *t++=*p++;
	*t='\0';
	++p;
	if (closing)
	  s=pop(s);
	else
	  {
	    tag_info **tp;
	    for (tp=tags; strcmp((*tp)->tag, the_tag) != 0; ++tp)
	      ;
	    s=do_it(s, *tp);
	  }
      }
    else if (is_space(*p))
      {
	context decoration=space_context(s->decoration);
	decorated_char *prev;
	if (decoration.tt)
	  output(m, *p, decoration);
	else if (m != meaning && is_space((prev=m-1)->c) && prev->decoration == decoration)
	  ;
	else
	  output(m, ' ', decoration);
	++p;
      }
    else
      output(m, *p++, s->decoration);
  free(s);
  meaning_len=m-meaning;
  m->c=0;
}


static
void
put(char c)
{
  putchar(c);
}


static
void
print_context(context const &decoration)
{
  put('[');
  if (decoration.b)
    put('B');
  if (decoration.em)
    put('E');
  if (decoration.i)
    put('I');
  if (decoration.s)
    put('S');
  if (decoration.tt)
    put('T');
  if (decoration.u > 0)
    put('0'+decoration.u);
  if (decoration.size == initial)
    put('?');
  else
    put('0'+decoration.size);
  static char colors[]={ 'r', 'g', 'b', 'c', 'm', 'y', 'k', 'w' };
  put(decoration.color == initial ? '?' : colors[decoration.color]);
  put(']');
}


static
void
print_meaning()
{
  for (decorated_char *m=meaning; m->c; ++m)
    {
      if (m == meaning || m->decoration != (m-1)->decoration)
	print_context(m->decoration);
      put(m->c);
    }
}


static int num_ts;


struct tag_sequence
{
  int refs;
  int pos;
  tag_info *tag;
  int closing;
  tag_sequence *next;

#ifdef CHECK_ALLOC
  tag_sequence()
  {
    ++num_ts;
  }

  ~tag_sequence()
  {
    --num_ts;
    assert(num_ts >= 0);
  }
#endif
};


static
tag_sequence *
copy(tag_sequence *ts)
{
  if (ts)
    ++ts->refs;
  return (ts);
}


static
void
free(tag_sequence *ts)
{
  while (ts && !--ts->refs)
    {
      tag_sequence *next=ts->next;
      delete ts;
      ts=next;
    }
}


static
tag_sequence *
push(int pos, tag_info *tag, int closing, tag_sequence *ts)
{
  tag_sequence *result=new tag_sequence;
  result->refs=1;
  result->pos=pos;
  result->tag=tag;
  result->closing=closing;
  result->next=ts;
  assert(!ts || ts->refs > 0);
  return (result);
}


static int num_sol;


struct solution
{
  int refs;
  int cost;
  context_stack *s;
  tag_sequence *ts;

#ifdef CHECK_ALLOC
  solution()
  {
    ++num_sol;
  }

  ~solution()
  {
    --num_sol;
    assert(num_sol >= 0);
  }
#endif
};


static
solution *
copy(solution *sol)
{
  ++sol->refs;
  return (sol);
}


static
void
free(solution *sol)
{
  if (!--sol->refs)
    {
      free(sol->s);
      free(sol->ts);
      delete sol;
    }
}


static
solution *
cow(solution *sol)
{
  if (sol->refs == 1)
    return (sol);
  --sol->refs;
  solution *result=new solution;
  result->refs=1;
  result->cost=sol->cost;
  result->s=copy(sol->s);
  result->ts=copy(sol->ts);
  return (result);
}


static
solution *
pop(solution *sol, int pos)
{
  sol=cow(sol);
  sol->ts=push(pos, sol->s->tag, 1, sol->ts);
  sol->s=pop(sol->s);
  return (sol);
}


static
solution *
push(tag_info *tag, solution *sol, int pos)
{
  sol=cow(sol);
  sol->cost+=cost(tag);
  sol->s=do_it(sol->s, tag);
  sol->ts=push(pos, tag, 0, sol->ts);
  return (sol);
}


struct heap
{
  int max_size;
  int size;
  int sign;
  solution **hs;

  heap(int sign_);
};


heap::heap(int sign_) : max_size(128), size(0), sign(sign_), hs((solution **)malloc(max_size*sizeof(hs[0])))
{
  ;
}


static
void
check_heap(heap *h)
{
  for (int i=1; i < h->size; ++i)
    {
      int p=(i-1)/2;
      assert(h->hs[p]->cost*h->sign >= h->hs[i]->cost*h->sign);
    }
}


static
void
add(solution *sol, heap *h)
{
  if (h->size == h->max_size)
    {
      h->max_size*=2;
      h->hs=(solution **)realloc(h->hs, h->max_size*sizeof(h->hs[0]));
    }
  int i=h->size++;
  while (i > 0)
    {
      int p=(i-1)/2;
      if (h->hs[p]->cost*h->sign > sol->cost*h->sign)
	break;
      else
	{
	  h->hs[i]=h->hs[p];
	  i=p;
	}
    }
  h->hs[i]=sol;
}


static
solution *
remove(heap *h)
{
  if (h->size == 0)
    return (0);
  solution *result=h->hs[0];
  if (--h->size)
    {
      solution *s=h->hs[h->size];
      int i=0;
      int c;
      while ((c=2*i+1) < h->size)
	{
	  if (c+1 < h->size && h->hs[c+1]->cost*h->sign > h->hs[c]->cost*h->sign)
	    ++c;
	  if (s->cost*h->sign > h->hs[c]->cost*h->sign)
	    break;
	  else
	    {
	      h->hs[i]=h->hs[c];
	      i=c;
	    }
	}
      h->hs[i]=s;
    }
  return (result);
}


static
int
included(solution *sol, heap *h)
{
  for (int i=0; i < h->size; ++i)
    if (h->hs[i]->s == sol->s)
      return (1);
  return (0);
}


static int max_beam_size=120;
static heap beam(1);
static heap candidates(-1);


static
void
print_beam()
{
  for (int i=0; i < beam.size; ++i)
    {
      solution *sol=beam.hs[i];
      printf("cost %d, ", sol->cost);
      print_context(sol->s->decoration);
      for (context_stack *s=sol->s; s->next; s=s->next)
	printf(", %s", s->tag->tag);
      put('\n');
    }
  put('\n');
  fflush(stdout);
}


static
int
compare_costs(void const *p1, void const *p2)
{
  solution *sol1=*(solution **)p1;
  solution *sol2=*(solution **)p2;
  return (sol1->cost-sol2->cost);
}


static
void
step(int pos)
{
  decorated_char m=meaning[pos];
  //  print_beam();
  candidates.size=0;
  for (int i=0; i < beam.size; ++i)
    {
      solution *sol=beam.hs[i];
      while (sol->s)
	{
	  if (included(sol, &candidates))
	    break;
	  add(copy(sol), &candidates);
	  sol=pop(sol, pos);
	}
      free(sol);
    }
  beam.size=0;
  int lowest_cost=999999999;
  while (candidates.size)
    {
      solution *sol=remove(&candidates);
      if (beam.size == max_beam_size && sol->cost > beam.hs[0]->cost && sol->cost > lowest_cost)
	free(sol);
      else
	{
	  context decoration=is_space(m.c) ? space_context(sol->s->decoration) : sol->s->decoration;
	  if (m.decoration == decoration)
	    {
	      if (lowest_cost > sol->cost)
		lowest_cost=sol->cost;
	      if (beam.size == max_beam_size)
		if (beam.hs[0]->cost == lowest_cost && max_beam_size < 120)
		  ++max_beam_size;
		else
		  free(remove(&beam));
	      add(sol, &beam);
	    }
	  else if ((m.decoration.size != decoration.size && m.decoration.size == initial) ||
		   (!is_space(m.c) && m.decoration.color != decoration.color && m.decoration.color == initial))
	    free(sol);
	  else
	    {
	      if (m.decoration.size != decoration.size)
		add(push(size_tags[m.decoration.size], copy(sol), pos), &candidates);
	      if (m.decoration.color != decoration.color && m.decoration.color != initial)
		add(push(color_tags[m.decoration.color], copy(sol), pos), &candidates);
	      if (m.decoration.b != decoration.b)
		if (!decoration.b)
		  add(push(&tB, copy(sol), pos), &candidates);
		else
		  add(push(&tPL, copy(sol), pos), &candidates);
	      if (m.decoration.em != decoration.em && !decoration.s)
		add(push(&tEM, copy(sol), pos), &candidates);
	      if (m.decoration.i != decoration.i)
		if (!decoration.i)
		  add(push(&tI, copy(sol), pos), &candidates);
		else
		  add(push(&tPL, copy(sol), pos), &candidates);
	      if (m.decoration.s != decoration.s)
		if (!decoration.s)
		  add(push(&tS, copy(sol), pos), &candidates);
		else
		  add(push(&tPL, copy(sol), pos), &candidates);
	      if (m.decoration.tt != decoration.tt)
		if (!decoration.tt)
		  add(push(&tTT, copy(sol), pos), &candidates);
		else
		  add(push(&tPL, copy(sol), pos), &candidates);
	      if (m.decoration.u != decoration.u)
		if (m.decoration.u > decoration.u)
		  add(push(&tU, copy(sol), pos), &candidates);
		else
		  add(push(&tPL, copy(sol), pos), &candidates);
	      free(sol);
	    }
	}
    }
  assert(beam.size);
  qsort(beam.hs, beam.size, sizeof(beam.hs[0]), compare_costs);
  for (int i=0; i < beam.size; ++i)
    {
      solution *soli=beam.hs[i];
      for (int j=i+1; j < beam.size; ++j)
	{
	  solution *solj=beam.hs[j];
	  int i_dominates_j=1;
	  context_stack *si=soli->s;
	  context_stack *sj=solj->s;
	  while (si && sj && si != sj)
	    if (si->tag != sj->tag)
	      {
		i_dominates_j=0;
		break;
	      }
	    else
	      {
		si=si->next;
		sj=sj->next;
	      }
	  if (i_dominates_j)
	    {
	      free(solj);
	      for (int k=j; k+1 < beam.size; ++k)
		beam.hs[k]=beam.hs[k+1];
	      --beam.size;
	      --j;
	    }
	}
    }
}


static double stop_time;


static
solution *
synthesize()
{
  solution *sol=new solution;
  sol->refs=1;
  sol->cost=0;
  sol->s=do_it((context_stack *)0, &tinitialize);
  sol->ts=0;
  add(sol, &beam);
  context last_decoration;
  initialize(last_decoration);
  int last_was_space=0;
  int total_steps=0;
  int pos;
  for (pos=0; meaning[pos].c; ++pos)
    {
      if (last_was_space && is_space(meaning[pos].c) && meaning[pos].decoration == last_decoration)
	continue;
      if (!last_was_space &&
	  (meaning[pos].decoration == last_decoration ||
	   (is_space(meaning[pos].c) && meaning[pos].decoration == space_context(last_decoration))))
	continue;
      ++total_steps;
      last_was_space=is_space(meaning[pos].c);
      last_decoration=meaning[pos].decoration;
    }
  initialize(last_decoration);
  last_was_space=0;
  int const steps_between_checks=1000;
  int steps_done=0;
  double start_time=now();
  double last_start_time=start_time;
  int steps_til_check=steps_between_checks;
  for (pos=0; meaning[pos].c; ++pos)
    {
      if (last_was_space && is_space(meaning[pos].c) && meaning[pos].decoration == last_decoration)
	continue;
      if (!last_was_space &&
	  (meaning[pos].decoration == last_decoration ||
	   (is_space(meaning[pos].c) && meaning[pos].decoration == space_context(last_decoration))))
	continue;
      step(pos);
      ++steps_done;
      last_was_space=is_space(meaning[pos].c);
      last_decoration=meaning[pos].decoration;
      if (--steps_til_check == 0)
	{
	  double current_time=now();
	  if (current_time > stop_time+10)
	    return (0);
	  double required_step_rate=(total_steps-steps_done)/(stop_time-current_time);
	  double last_step_rate=steps_between_checks/(current_time-last_start_time);
	  double ratio=0.8*last_step_rate/required_step_rate;
	  if (ratio > 1.2)
	    ratio=1.2;
	  else if (ratio < 0.8)
	    ratio=0.8;
	  max_beam_size=(int)(max_beam_size*ratio);
	  if (max_beam_size > 120)
	    max_beam_size=120;
	  else if (max_beam_size < 10)
	    max_beam_size=10;
	  last_start_time=current_time;
	  steps_til_check=steps_between_checks;
	}
    }
  sol=beam.hs[0];
  for (int i=1; i < beam.size; ++i)
    free(beam.hs[i]);
  while (sol->s->next)
    sol=pop(sol, pos);
  sol->cost+=pos;
  return (sol);
}


static
tag_sequence *
reverse(tag_sequence *ts)
{
  tag_sequence *result=0;
  while (ts)
    {
      tag_sequence *next=ts->next;
      ts->next=result;
      result=ts;
      ts=next;
    }
  return (result);
}


static
tag_sequence *
print_tags(tag_sequence *ts, int pos)
{
  while (ts && ts->pos == pos)
    {
      put('<');
      if (ts->closing)
	put('/');
      fputs(ts->tag->tag, stdout);
      put('>');
      tag_sequence *next=ts->next;
      delete ts;
      ts=next;
    }
  return (ts);
}


static
void
print_solution(solution *sol)
{
  tag_sequence *ts=reverse(sol->ts);
  int pos;
  for (pos=0; meaning[pos].c; ++pos)
    {
      ts=print_tags(ts, pos);
      put(meaning[pos].c);
    }
  print_tags(ts, pos);
  sol->ts=0;
  free(sol);
}


int
main(int argc, char **argv)
{
  snarf();
  compute_meaning(input);
  if (strcmp(*argv, "meaning") == 0)
    print_meaning();
  else
    {
      if (argc > 1)
	stop_time=now()+atoi(argv[1])-20;
      else
	stop_time=now()+160;
      solution *sol=synthesize();
      if (!sol || (unsigned)sol->cost >= strlen(input))
	fputs(input, stdout);
      else
	print_solution(sol);
      if (num_s+num_ts+num_sol)
	fprintf(stderr, "%d context stacks, %d tag sequences, %d solutions\n", num_s, num_ts, num_sol);
    }
  return (0);
}
